import errno
from functools import partial
import json
import os
import subprocess


################################################################################
# The following chunk of code for querying pkg-config is loosely based on code
# from the cffi package, which is licensed as follows:
#
#    The MIT License
#
#    Permission is hereby granted, free of charge, to any person
#    obtaining a copy of this software and associated documentation
#    files (the "Software"), to deal in the Software without
#    restriction, including without limitation the rights to use,
#    copy, modify, merge, publish, distribute, sublicense, and/or
#    sell copies of the Software, and to permit persons to whom the
#    Software is furnished to do so, subject to the following conditions:
#
#    The above copyright notice and this permission notice shall be included
#    in all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#    DEALINGS IN THE SOFTWARE.

def _ask_pkg_config(lib, option, result_prefix='', sysroot=False, default=None):
    pkg_config = os.environ.get('PKG_CONFIG', 'pkg-config')
    try:
        output = subprocess.check_output([pkg_config, option, lib])
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise
    except subprocess.CalledProcessError:
        pass
    else:
        res = output.decode().strip().split()

        # '-I/usr/...' -> '/usr/...'
        for x in res:
            assert x.startswith(result_prefix)
        res = [x[len(result_prefix):] for x in res]

        if sysroot:
            sysroot = os.environ.get('PKG_CONFIG_SYSROOT_DIR', '')
        if sysroot:
            # old versions of pkg-config don't support this env var,
            # so here we emulate its effect if needed
            res = [path if path.startswith(sysroot) else sysroot + path
                   for path in res]

        return res
    return [] if default is None else default


def get_pkg_info(name):
    ask = partial(_ask_pkg_config, name)
    return {
        'libraries': ask('--libs-only-l', '-l', default=[name]),
        'include_dirs': ask('--cflags-only-I', '-I', sysroot=True),
        'library_dirs': ask('--libs-only-L', '-L', sysroot=True),
        'extra_compile_args': ask('--cflags-only-other'),
        'extra_link_args': ask('--libs-only-other'),
        'runtime_library_dirs': ask('--variable=libdir'),
    }


################################################################################
### Get FLANN info

_flann_info = None
def get_flann_info(flann_dir=None, use_cache=True):
    global _flann_info
    if _flann_info is not None and use_cache:
        return _flann_info

    if flann_dir is None:
        flann_dir = os.environ.get('FLANN_DIR')

    if flann_dir is None:
        pth = os.path.join(os.path.dirname(__file__), 'flann_config.json')
        try:
            with open(pth, 'r') as f:
                meta = json.load(f)
        except IOError:
            pass
        else:
            flann_dir = meta.get('FLANN_DIR')

    if flann_dir:
        pre = partial(os.path.join, flann_dir)
        lib_dirs = [pre('lib')]
        if os.path.isdir(pre('lib64')) and sys.maxsize > 2**32:
            lib_dirs.append(pre('lib64'))

        _flann_info = {
            'libraries': ['flann', 'flann_cpp'],
            'include_dirs': [pre('include')],
            'library_dirs': lib_dirs,
            'extra_compile_args': [],
            'extra_link_args': [],
            'runtime_library_dirs': lib_dirs,
        }
    else:
        _flann_info = get_pkg_info('flann')
        if _flann_info['libraries'] == ['flann_cpp']:
            import warnings
            warnings.warn("It looks like you're using an old version of FLANN "
                          "with a bug in its pkg-config settings. Trying to "
                          "work around this, but you'd be better off setting "
                          "FLANN_DIR as described in the README, or updating "
                          "to FLANN 1.9 or higher.")
            _flann_info['libraries'].insert(0, 'flann')

    if os.name == 'nt':
        _flann_info['runtime_library_dirs'] = []
        # flann_cpp.lib doesn't exist for some reason
        _flann_info['libraries'].remove('flann_cpp')

    return _flann_info



import sys

# versioneer: get __version__ attribute
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions


# A bit of a hack, following numpy/sklearn: figure out whether we're being
# imported in the setup phase and so shouldn't try to import the compiled
# extensions.
try:
    __CYFLANN_SETUP__
except NameError:
    __CYFLANN_SETUP__ = False

if __CYFLANN_SETUP__:
    sys.stderr.write("Partial import of cyflann during the build process.\n")
else:
    from . import index
    from .index import FLANNParameters, FLANNIndex, set_distance_type

from . import flann_info, extensions
from .flann_info import get_flann_info
from .extensions import FLANNExtension

# A test function, if we have nose.
try:
    from numpy.testing import nosetester
    test = nosetester.NoseTester().test
    del nosetester
except ImportError:
    pass



'''
Helpers for consumers to compile FLANN extensions.
'''
import os
import subprocess
import sys

try:
    from setuptools.extension import Extension
    from setuptools.command.build_ext import build_ext
except ImportError:
    from distutils.extension import Extension
    from distutils.command.build_ext import build_ext

from .flann_info import get_flann_info


class FLANNExtension(Extension):
    def __init__(self, name, sources, **kw):
        for key, vals in get_flann_info().items():
            kw[key] = l = kw.get(key) or []
            for v in vals:
                if v not in l:
                    l.append(v)

        Extension.__init__(self, name, sources, **kw)


# Used to use this to do install_name_tool munging on OSX, but that was only
# needed for old conda and shouldn't be necessary anymore. Still here for
# backwards compatability.
class build_ext_flann(build_ext):
    pass



from copy import deepcopy
import os
import sys
import versioneer

if sys.version_info[0] < 3:
    import __builtin__ as builtins
else:
    import builtins

try:
    import numpy
except ImportError as e:
    raise ImportError("cyflann requires numpy to be installed:\n{}".format(e))
    # Don't do this in the setup() requirements, because otherwise pip and
    # friends get too eager about updating numpy.

try:
    from setuptools import setup, Command
    from setuptools.extension import Extension
    from setuptools.command.build_ext import build_ext
    from pprint import pprint

    class PrintInfoCommand(Command):
        user_options = []

        def initialize_options(self):
            pass

        def finalize_options(self):
            pass

        def run(self):
            global flann_info
            pprint(flann_info)

except ImportError:
    from distutils.core import setup
    from distutils.extension import Extension
    from distutils.command.build_ext import build_ext

    class PrintInfoCommand():
        pass

# A hack, following sklearn: set a global variable so that cyflann.__init__
# knows not to try to import compiled extensions that aren't built yet during
# the setup process.
builtins.__CYFLANN_SETUP__ = True
from cyflann.flann_info import get_flann_info
flann_info = get_flann_info()


try:
    from Cython.Build import cythonize
except ImportError:
    src_path = os.path.join(os.path.dirname(__file__), 'cyflann')
    if not os.path.exists(os.path.join(src_path, 'index.c')):
        msg = "index extension needs to be compiled but cython isn't available"
        raise ImportError(msg)
    ext_modules = [
        Extension("cyflann.index", ["cyflann/index.c"]),
    ]
else:
    ext_modules = cythonize([
        Extension("cyflann.index", ["cyflann/index.pyx"],
                  depends=["cyflann/index.pxd", "cyflann/flann.pxd"])
    ])

for ext in ext_modules:
    ext.__dict__.update(deepcopy(flann_info))
    ext.include_dirs.append(numpy.get_include())

cmdclass = versioneer.get_cmdclass()
cmdclass['flann_info'] = PrintInfoCommand

setup(
    name='cyflann',
    version=versioneer.get_version(),
    author='Dougal J. Sutherland',
    author_email='dougal@gmail.com',
    packages=['cyflann', 'cyflann.tests'],
    package_data={'cyflann': ['*.pyx', '*.pxd']},
    url='https://github.com/dougalsutherland/cyflann',
    description='A Cython-based interface to the FLANN nearest neighbors '
                'library.',
    long_description=open('README.rst').read(),
    license='BSD 3-clause',
    ext_modules=ext_modules,
    cmdclass=cmdclass,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Cython",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    zip_safe=False,  # not unsafe but no point, since it's just a c ext
)



from .flann cimport FLANNParameters as Parameters, flann_index_t as index_t

cdef class FLANNParameters:
    cdef Parameters _this

cdef class FLANNIndex:
    cdef index_t _this
    cdef float[:, ::1] _data

    cdef public float speedup
    cdef public FLANNParameters params
    cdef public object _rn_gen

    cdef void _free_index(self) nogil

    cpdef _check_array(self, array, int dim=?)
    cdef _ensure_random_seed(self, kwargs)

    cdef void _nn(self, float[:, ::1] pts, float[:, ::1] qpts,
                  int num_neighbors, int[:, ::1] idx, float[:, ::1] dists) nogil
    cdef void _build_index(self, float[:, ::1] pts) nogil

    cpdef save_index(self, bytes filename)
    cdef void _load_index(self, bytes filename, float[:, ::1] pts)

    cdef void _nn_index(self, float[:, ::1] qpts, int num_neighbors,
                        int[:, ::1] idx, float[:, ::1] dists) nogil
    cdef int _nn_radius(self, float[:] query, float radius, int max_nn,
                        int[:] idx, float[:] dists) nogil

    cdef int _hierarchical_kmeans(self, float[:, ::1] pts, int num_clusters,
                                  float[:, ::1] result) nogil



from functools import partial

import numpy as np
import pyflann
import pytest
import cyflann


@pytest.mark.parametrize("dim", [1, 4, 10])
@pytest.mark.parametrize("k", [1, 2, 5])
def test_normal(dim, k):
    data = np.random.normal(scale=100, size=(1000, dim))
    query = np.random.normal(scale=100, size=(100, dim))

    py = pyflann.FLANN(algorithm='kdtree_single')
    cy = cyflann.FLANNIndex(algorithm='kdtree_single')

    py.build_index(data)
    cy.build_index(data)

    py_ids, py_dists = py.nn_index(query, k)
    cy_ids, cy_dists = cy.nn_index(query, k)

    assert np.all(py_ids == cy_ids), \
           "{}/{} different".format(np.sum(py_ids != cy_ids), py_ids.size)
    assert np.allclose(py_dists, cy_dists, atol=1e-5, rtol=1e-4), \
           "max distance {}".format(np.abs(py_dists - cy_dists).max())
